@@ -1,4 +1,7 @@
-"""Database management
+"""Pure-Python database engine
+
+Licence : BSD
+Author : Pierre Quentel (pierre.quentel@gmail.com)
 
 Access by list comprehension or generator expression or select
 
@@ -68,9 +71,19 @@ Version 1.4
 floats)
 - move methods to manage db information to new module buzhug_info.py
 - add negative indexing. db[-1] returns the last inserted record
+Version 1.5
+- add thread-safe version class TS_Base
+- introduce syntax db(key1=value2[,key2=value2...])
+- allow an iterable of records for update
+- fix bug in Base.has_key(), UnicodeFile.from_block()
+Version 1.6
+- minor bug fix in conversion_float.py
+- add licence text
+- fix bug in thread-safe version
 """
 
 import os
+import threading
 import cStringIO
 import itertools
 import token
@@ -93,7 +106,7 @@ from buzhug_files import *
 import buzhug_algos
 import buzhug_info
 
-version = "1.3"
+version = "1.6"
 
 # ------------------------
 # The following classes are used to modelize the objects in the
@@ -137,7 +150,7 @@ class Record(list):
             if not isinstance(getattr(self,k),unicode):
                 elts.append('%s:%s' %(k,getattr(self,k)))
             else:
-                elts.append(('%s:%s' %(k,getattr(self,k))).encode('latin-1'))
+                elts.append(('%s:%s' %(k,getattr(self,k))).encode('utf-8'))
         return '<' + ' '.join(elts) +'>'
 
     def update(self,**kw):
@@ -296,6 +309,7 @@ class TimeFormatError(Exception):
     """Raised if an invalid time format is provided to set_string_format"""
     pass
 
+
 class Base:
 
     BLOCKSIZE = 131072
@@ -306,7 +320,7 @@ class Base:
             (date,DateFile),(datetime,DateTimeFile), (dtime, TimeFile),
             (bool,BooleanFile)]
 
-    def __init__(self,basename):
+    def __init__(self,basename,thread_safe=False):
         self.name = self.__name__ = basename
         self.types = {} # key = data class name, value = data class
         self.file_types = {} # key = data class, value = file class
@@ -371,12 +385,18 @@ class Base:
                 else:
                     raise IOError,"Base %s already exists" %self.name
             else:
-                raise IOError,"Directory %s already exists" %self.name
+                if mode != 'open':
+                    raise IOError,"Directory %s already exists" %self.name
+                else:
+                    raise IOError,"Mode 'open' : " \
+                        "Directory %s already exists but no info file found" \
+                        %self.name
+
         self.field_names = [ f[0] for f in fields ]
         self.fields = dict([(f[0],f[1]) for f in fields])
         # set general info about field definition
         buzhug_info.set_info(self,fields)
-        
+
         self.field_names = ['__id__','__version__'] + self.field_names
         self.fields['__id__'] = int
         self.fields['__version__'] = int
@@ -704,6 +724,9 @@ class Base:
         res,names = self._select(names,request,**args)
         return ResultSet(names,res.values())
 
+    def __call__(self,**kw):
+        return self.select_for_update(**kw)
+
     def _select(self,_names,_request,**args):
         """Private method that performs actual selection
         The field files are browsed line by line. A test function is built
@@ -821,6 +844,11 @@ class Base:
         If only fixed length fields have changed, simply put the new values
         at the same position in field files
         Otherwise, remove existing record then insert the new version"""
+        if not isinstance(record,Record) \
+            and isinstance(record,(list,tuple)):
+            for rec in record:
+                self.update(rec,**kw)
+            return
         only_fixed_length = True
         if '__id__' in kw.keys():
             raise NameError,"Can't update __id__"
@@ -987,11 +1015,15 @@ class Base:
         return self._pos._count
 
     def has_key(self,num):
-        p_block = self._pos.get_block_at(num)
-        if not p_block or p_block[0] == '#':
+        # first find the line in position file
+        block_pos = self._id_pos.get_block_at_pos(5*num)
+        if not block_pos or block_pos[0] == '#':
             return False
         return True
 
+    def __contains__(self,num):
+        return self.has_key(num)
+
     def keys(self):
         return [ r.__id__ for r in self.select(['__id__']) ]
 
@@ -1005,3 +1037,125 @@ class Base:
             if record[0][0] != "#":
                 r = self._full_rec(record)
                 yield r
+
+# thread-safe base ; copied from the logging module
+_lock = None
+
+def _acquireLock():
+    """
+    Acquire the module-level lock for serializing access to shared data.
+
+    This should be released with _releaseLock().
+    """
+    global _lock
+    if (not _lock):
+        _lock = threading.RLock()
+    if _lock:
+        _lock.acquire()
+
+def _releaseLock():
+    """
+    Release the module-level lock acquired by calling _acquireLock().
+    """
+    if _lock:
+        _lock.release()
+
+class TS_Base(Base):
+
+    def create(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.create(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def open(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.open(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def close(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.close(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def destroy(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.destroy(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def set_default(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.set_default(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def insert(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.insert(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def update(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.update(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def delete(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.delete(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def cleanup(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.cleanup(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def commit(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.commit(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def add_field(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.add_field(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    def drop_field(self,*args,**kw):
+        _acquireLock()
+        try:
+            res = Base.drop_field(self,*args,**kw)
+        finally:
+            _releaseLock()
+        return res
+
+    
\ No newline at end of file
